/*
 * Md5 brute forcer for CELL architecture
 *
 * Simon Marechal
 * Raboin Romain <romain.raboin at gmail.com> (2008)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <libspe.h>
#include <errno.h>
#include <sys/time.h>
#include <time.h>
#include <xyssl/md5.h>

#ifndef MD5_DIGEST_LENGTH
#define MD5_DIGEST_LENGTH 16
#endif

#define NB_BUF	6
#define PARA	4
#define NB_PWD	(4*PARA)
#define CHARSET 62



#define H(x) ( ((x&0xff)<<24) | ((x&0xff00)<<8) | ((x&0xff0000)>>8) | (x>>24 ) )
#define IDX(pwd, idx, bf) (((pwd) + (idx) * NB_PWD * PARA / 4 + (bf) * 64 * NB_PWD / 4))

unsigned int size;
unsigned char update[256];
unsigned long long gen;

extern spe_program_handle_t md5_body;


typedef struct {
	unsigned char	*pwd[4];
	int		mb;
} save_pwd_t;

save_pwd_t	pwd_list[NB_BUF];

void	init(void)
{
	unsigned int i;

	memset(update, 0, 256);
	update[0] = 'a';
	for (i = 'a'; i < 'z'; i++)
		update[i] = i + 1;
	update['z'] = 'A';
	for (i = 'A'; i < 'Z'; i++)
		update[i] = i + 1;
	update['Z'] = '0';
	for (i = '0'; i < '9'; i++)
		update[i] = i + 1;
	gen = 0;
}

unsigned char nb2asc(unsigned char x)
{
	if (x < 26)
		return x + 'a';
	if (x < 52)
		return x - 26 + 'A';
	return x - 52 + '0';
}


/* save current password for each mailbox of each spe */
void	save_pwd(unsigned char *pwd, unsigned int manage)
{
	save_pwd_t	*sp;

	sp = &pwd_list[manage];
	sp->mb++;
	if (sp->mb >= 4)
		sp->mb = 0;
	strncpy((char*)sp->pwd[sp->mb], (char*)pwd + 2, strlen((char*)pwd+2) - 1);
}

/* generate password for spe */
void    fill_buffer(unsigned char *buffer, unsigned int curbuf, unsigned char *pwd)
{
        unsigned int i, j, k;

        for (i = 0; i < PARA; i++) {
                k = 2;
                while ((pwd[k] = update[pwd[k]]) == 0) {
                        pwd[k] = update[0];
                        k++;
                        if (k > size) {
                                size = k;
                                pwd[k] = 0;
                                pwd[k + 1] = 0x80;
                                printf("upsize %s (%d)!\n", pwd + 2, size+1);
                        }
                }
                for (k = 0; k < PARA; k++) {
                        for(j = 0; j < size; j++)
                                ((unsigned int *)buffer)[IDX(i * PARA + k, j, curbuf)] = H(((unsigned int *)pwd)[j]);
                        ((unsigned int *)buffer)[IDX(i + k * PARA, 14, curbuf)] = (((size + 1) * 8));
                        ((unsigned int *)buffer)[IDX(i + k * PARA, 15, curbuf)] = 0;
                }
                gen++;
        }
}


/* compare hash */
int	cmphash(char *hash, char *tofind)
{
	int	i;
	int	c;
	char	tmp[4];

	for (i = 0; i < 4; i++)
		tmp[i] = hash[4 - i - 1];
	c = *(int *)tmp;
	if (c == *(int*)tofind)
		return 1;
	return 0;
}

/* brute force first char */
void	testpwd(int manage, char *tofind)
{
	int	r1, r2, r3, mb;
	char	*charset = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789";
	char	res[MD5_DIGEST_LENGTH];
	char	pwd[32];


	memset(pwd, 0, sizeof(pwd));
	pwd[0] = charset[0];
	pwd[1] = charset[0];
	for (mb = 0; mb < 4; mb++) {
		strcpy(pwd + 2, (char*)pwd_list[manage].pwd[mb]);
		pwd[size + 1] = 0;
		for (r3 = 0; r3 < strlen(charset); r3++) {
			for (r2 = 0; r2 < strlen(charset); r2++) {
				for (r1 = 0; r1 < strlen(charset); r1++) {
					pwd[0] = charset[r1];
					pwd[1] = charset[r2];
					pwd[2] = charset[r3];
					md5((unsigned char*)pwd, strlen((char*)pwd), (unsigned char*)res);
					if (cmphash(res, tofind) == 1) {
						printf("Hash found: %s\n", pwd);
						return;
					}
				}
			}
		}
	}
	printf("Not found\n");
}

/* convert hash from command line */
int	gethash(char *hash, char *tofind)
{
	char	tmp[9];
	int	i, j;

	tmp[8] = 0;
	if (strlen(hash) != 32)
		return 0;
	for (j = 0; j < 16; j += 4) {
		for (i = 0; i < 8; ) {
			tmp[i+1] = hash[j * 2 + 8 - i - 1];
			i++;
			tmp[i-1] = hash[j * 2 + 8 - i - 1];
			i++;
		}
		*(int*)&tofind[j] = strtoul(tmp, NULL, 16);
	}
	return 1;
}


int	main(int ac, char **av)
{
	char		tofind[MD5_DIGEST_LENGTH];
	unsigned char	buffer[64 * NB_PWD * NB_BUF]	__attribute__((aligned(16)));
	unsigned char	hashes[16 * NB_BUF]		__attribute__((aligned(16)));
	unsigned char	pwd[32];
	unsigned int	i, j, k;
	unsigned int	curbuf;
	unsigned int	lu;
	speid_t spe_id[NB_BUF];
	struct {
		unsigned char *buffer		__attribute__((aligned(16)));
		unsigned long long *hashes	__attribute__((aligned(16)));
		unsigned char *bf		__attribute__((aligned(16)));
		unsigned int ta[4]		__attribute__((aligned(16)));
		unsigned int tb[4]		__attribute__((aligned(16)));
		unsigned int tc[4]		__attribute__((aligned(16)));
		unsigned int td[4]		__attribute__((aligned(16)));
	} spe_struct[NB_BUF]			__attribute__((aligned(16)));
	unsigned long long tested = 0;
	unsigned char bf_buf[CHARSET * CHARSET * 4];


	if (ac != 2) {
		printf("%s <hash>\n", av[0]);
		return -1;
	}
	printf("Hash : %s\n", av[1]);
	if (gethash(av[1], tofind) == 0) {
		printf("Invalid argument\n");
		return -1;
	}

	for (i = 0; i < CHARSET; i++) {
		for (j = 0; j < CHARSET; j++) {
			for (k = 0; k < 4; k++) {
				bf_buf[(i + j * CHARSET) * 4 + 3] = nb2asc(i);
				bf_buf[(i + j * CHARSET) * 4 + 2] = nb2asc(j);
				bf_buf[(i + j * CHARSET) * 4 + 1] = 0;
				bf_buf[(i + j * CHARSET) * 4 + 0] = 0;
			}
		}
	}

	memset(buffer, 0, sizeof(buffer));
	memset(pwd, 0, sizeof(pwd));

	init();

	printf("got %d SPEs, using %d\n", spe_count_physical_spes(), NB_BUF);

	pwd[0] = 0;
	pwd[1] = 0;
	pwd[2] = nb2asc(0);
	pwd[3] = 0x80;
	size = 3;

	for (i = 0; i < NB_PWD; i++) {
		for (curbuf = 0; curbuf < NB_BUF; curbuf++) {
			((unsigned int *)buffer)[IDX(i, 14, curbuf)] = (((size) * 8));
		}
	}


	memset(pwd_list, 0, sizeof(pwd_list));
	/* Init spe and create spe thread */
	for (i = 0; i < NB_BUF; i++) {
		for (j = 0; j < 4; j++)
			pwd_list[i].pwd[j] = malloc(sizeof(pwd));
		pwd_list[i].mb = 0;
		spe_struct[i].buffer = buffer + 64 * NB_PWD * i;
		spe_struct[i].hashes = (unsigned long long *)(hashes + 16 * i);
		spe_struct[i].bf = bf_buf;
		for (j = 0; j < 4; j++)
			spe_struct[i].ta[j] = *(int*)&tofind[0];
		for (j = 0; j < 4; j++)
			spe_struct[i].tb[j] = *(int*)&tofind[4];
		for (j = 0; j < 4; j++)
			spe_struct[i].tc[j] = *(int*)&tofind[8];
		for (j = 0; j < 4; j++)
			spe_struct[i].td[j] = *(int*)&tofind[12];

		
		spe_id[i] = spe_create_thread(0, &md5_body, &spe_struct[i], NULL, -1, 0);
		fill_buffer(buffer, i, pwd);
		spe_write_in_mbox(spe_id[i], 1);
		spe_write_in_mbox(spe_id[i], 0);
	}

	struct timeval tv1;
	struct timeval tv2;

	gettimeofday(&tv1, NULL);

	/* crack loop */
	int manage;
	while (1) {
		manage = -1;
		for (i = 0; i < NB_BUF; i++) {
			if ((spe_stat_out_mbox(spe_id[i]))) {
				manage = i;
				break;
			}
		}
		if (manage == -1)
			continue;

		lu = spe_read_out_mbox(spe_id[manage]);

		tested += NB_PWD / PARA;
		if ((spe_struct[manage].hashes)[0] || (spe_struct[manage].hashes)[1]) {
			printf("Hash found in spe : %d\n", manage);
			testpwd(manage, tofind);
 			break;
		}
		save_pwd(pwd, manage);
		if (((tested / (NB_PWD / PARA)) % 10000) == 0)
			spe_write_in_mbox(spe_id[i], 1);
		else
			spe_write_in_mbox(spe_id[i], 0);
		fill_buffer(buffer, manage, pwd);

	}

	gettimeofday(&tv2, NULL);
	printf("%lldM passwords tested\n", (tested * CHARSET * CHARSET / 1000000));

	double crk_avg;

	crk_avg = ((double) tv2.tv_sec - tv1.tv_sec) + ((double) tv2.tv_usec - (double) tv1.tv_usec) / 1000000;
	printf("%fM pwd/s!\n\n", ((double)tested * CHARSET * CHARSET / 1000000) / crk_avg);
	return 0;
}
